#!/usr/bin/env python3

import numpy as np
import textwrap

def selav_array(name,arr,fmt='%e',n=10):
	strings = [ fmt%x if i%n else ('\n'+fmt)%x for i,x in enumerate(arr) ]
	values = ', '.join(strings)
	return(f'{name} = [{values}];')

def lattice2selav(fname_lattice, fname_beta, ofd):

	# read beta data
	beta_z, beta_x, beta_y = np.loadtxt(fname_beta).transpose()
	betalines = []
	betalines.append(selav_array('beta_data_z',beta_z,fmt='%0.3f'))
	betalines.append(selav_array('beta_data_x',beta_x,fmt='%0.3f'))
	betalines.append(selav_array('beta_data_y',beta_y,fmt='%0.3f'))


	variables = {}
	def addvar(name,val,com=''):
		variables[name] = (val,com)

	addvar('dryrun',            1,       'no propagation; only calculate lattice properties')
	addvar('center',            'AUTO')
	addvar('box',               'AREA')
	addvar('energy',            5e6,     'initial energy [eV]')
	addvar('emittance',         1e-6,    'transverse emittance [mm mrad]')
	addvar('spacecharge',       1,       'whether to include spacecharge')
	addvar('csr',               1,       'whether to include csr')
	addvar('csr_filter',        0,       '0: no filter, 1: low-pass, 2: gaussian')
	addvar('csr_filterwidth',   10.,     'in units of the Nyquist frequency')
	addvar('csr_mode',          5,       'csr mode to use; see manual')
	addvar('csr_transient',     0,       'include transient csr term (T2)')
	addvar('driftres',          1,       'drifts will be split in slices of this length [m]')
	addvar('nsplit_cavity',     5,       'cavities are split into this number of slices')
	addvar('nsplit_dipole',     10,      'dipoles ...')
	addvar('nsplit_quadrupole', 5,       'quadrupoles ...')
	addvar('cavity_mode',       0,       '0: non-linear, 1: linear')
	addvar('dipole_mode',       0,       '0: non-linear, 1: linear')

	# parse lattice file
	lattice = []
	with open(fname_lattice,"r") as f:
		for line in f.readlines():
			line = line.strip()
			if(line=='' or line[0]=="#"):
				continue
			line = line.split()
			keyw = line[0]
			name = line[1]
			z0 = float(line[2])
			z1 = float(line[3])
			if(keyw=='cavity'):
				deltae = float(line[4]) if len(line)>4 else 0.0
				freq   = float(line[5]) if len(line)>5 else 0.0
				phase  = float(line[6]) if len(line)>6 else 0.0
				params = [deltae, freq, phase]
			elif(keyw=='dipole'):
				k0orfactor = float(line[4]) if len(line)>4 else 0.0
				params = [k0orfactor]
			elif(keyw=='quadrupole'):
				k1	= float(line[4]) if len(line)>4 else 0.0
				params = [k1]

			lattice.append([keyw,name,z0,z1,params])

	# prepare chicanes
	dipolecount = {}
	dipolepos = {}
	for e in lattice:
		keyw, name, z0, z1, *_ = e
		if(not keyw=='dipole'):
			continue
		dipolecount.setdefault(name, 0)
		dipolepos.setdefault(name, [])

		dipolecount[name] += 1
		dipolepos[name].append([z0,z1])

	chicanes = {}
	for d in dipolepos:
		lbs = [z1-z0 for z0,z1 in dipolepos[d]]
		lds = [b-a for [_,a],[b,_] in zip(dipolepos[d][:-1],dipolepos[d][1:])]
		chicanes[d] = [lbs[-1], lds[-1]]
	is_c_chicane = {k: dipolecount[k]==4 for k in dipolecount} # TODO add better check to detect c-chicanes
	print(chicanes)




	latlines = []
	latlines.append('''// output of initial psd
elementname = "initial";
prefix = format("%04g-%04g-initial", nelement, substep);
eval(hook_post_element);
step=step+1;
'''
	)

	def make_step(z0, z1, k0, k1, mapstr, ename='unknown', nsteps='5', pre='', post=''):
		mapstr=mapstr.strip().replace('\t','')
		mapstr=textwrap.indent(mapstr,'\t')
		latlines.append(f'''
elementname = "{ename:s}";
z0 = {z0}; z1 = {z1}; k0 = {k0}; k1 = {k1}; N = {nsteps:s};\
{pre}
calc_map = {{
{mapstr:s}
}};
eval(element_loop);\
{post}
'''
		)

	z0prev=0 # start of previous element
	z1prev=0 # end of ...
	for e in lattice:
		keyw, name, z0, z1, params = e

		if z0<z1prev:
			sys.exit(f'error: element {name} not in order')

		# DRIFTS
		make_step(z1prev, z0, 0.0, 0.0,
			"M = map_identity();",
			ename='drift',
			nsteps = 'floor((z1-z0)/driftres+1)'
		)

		# CAVITY
		if(keyw=='cavity'):
			deltae, freq, phase = params

			addvar(f'{name}_deltae', deltae, 'energy gain [eV]')
			addvar(f'{name}_freq'  , freq,   'frequency [Hz]')
			addvar(f'{name}_phase' , phase,  'phase [rad]')

			make_step(z0, z1, 0.0, 0.0, f'''
				M = map_cavity(ampl={name}_deltae/cos({name}_phase)/N, freq={name}_freq, phase={name}_phase, mode=cavity_mode);
				energy = energy + {name}_deltae/N;''',
				ename  = f'{name}',
				nsteps = 'nsplit_cavity'
			)

		# DIPOLE
		elif(keyw=='dipole'):
			k0orfactor, = params

			if f'{name}_k0' in variables:
				factor = k0orfactor
			else:
				addvar(f'{name}_k0', k0orfactor, 'dipole strength [1/m]')
				factor = 1
				if(name in chicanes):
					lB,lD = chicanes[name]
					addvar(f'{name}_lB', lB, 'effective dipole length [m] (do not change)')
					if(is_c_chicane[name]):
						addvar(f'{name}_lD', lD, 'drift length between dipoles 1,2 and 3,4 [m] (do not change)')


			if(is_c_chicane[name] and dipolecount[name]==4):
				addvar(f'{name}_full', 1, '')
				latlines.append(f'''if({name}_full) {{
angle_{name} = asin({name}_k0 * {name}_lB);
if(dipole_mode==0) {{
R566_{name}  = -1/2*chicanecoef(2, angle_{name}, {name}_lB, {name}_lD) / 4 / nsplit_dipole;
R5666_{name} = -1/6*chicanecoef(3, angle_{name}, {name}_lB, {name}_lD) / 4 / nsplit_dipole;
}};
''')
			dipolecount[name]-=1

			make_step(z0, z1, f'{name}_k0 * {factor}', 0.0, f'''
				R566_{name}  := 0.0; // default to 0
				R5666_{name} := 0.0; //
				M = map_driftpoly([0, R56/energy, R566_{name}/energy**2, R5666_{name}/energy**3]);
				eval(add_csr);''',
				ename  = f'{name}',
				pre    = '\nR = transfer4d(R, 0.0, angle, k0, 0.0);',
				post   = '\nR = transfer4d(R, 0.0, angle, k0, 0.0);',
				nsteps = 'nsplit_dipole'
			)

			if(is_c_chicane[name] and dipolecount[name]==0):
				# last dipole of the family
				latlines.append('} else {')
				make_step(dipolepos[name][0][0], dipolepos[name][-1][-1], 0.0, 0.0, f'''
					M = map_chicanec(alpha=asin({name}_k0 * {name}_lB), lB={name}_lB, lD={name}_lD, energy=energy, mode=dipole_mode);''',
					ename  = f'{name}',
					nsteps = '1'
				)
				latlines.append('};')

		# QUADRUPOLE
		elif(keyw=='quadrupole'):
			k1,	= params

			addvar(f'{name}_k1', k1, 'quadrupole strength [1/m^2]')

			make_step(z0, z1, 0.0, f'{name}_k1', f'''
				M = map_driftpoly([0, R56/energy]);''',
				ename  = f'{name}',
				nsteps = 'nsplit_quadrupole'
			)

		z1prev = z1




	varlines = []
	varlines.append('// default values for variables')
	for var in variables:
		val,com = variables[var]
		if type(val) is float:
			vline = f'{var:20s} := {val:10g};'
		if type(val) is int:
			vline = f'{var:20s} := {val:10d};'
		if type(val) is str:
			vline = f'{var:20s} := "{val:s}";'
		if(com):
			vline += f' // {com:s}'
		varlines.append(vline)


	macrodefs='''\

if( defined(psi)==0 ) { // set dryrun if psi is not defined
	print("warning: psi not defined, doing dryrun");
	dryrun = 1;
};

propagation = {
	psi = propagate(psi, M, center=center, box=box);
};

calc_dispersion = {
	Rprev = R;
	R = transfer4d(R, ds, 0.0, k0, k1);
	R56      = R[11]-Rprev[11];
	R56total = R[11];
};

calc_beamsize = {
	gamma = energy/510.99895e3;
	betax = average(z-dz, z, beta_data_z, beta_data_x);
	betay = average(z-dz, z, beta_data_z, beta_data_y);
	beamsize = (sqrt(betax)+sqrt(betay))/2 * sqrt(emittance/gamma);
};

add_spacecharge = {
	if(spacecharge) {
		SC = map_spacecharge(psi,
			length=ds,
			beamsize=beamsize,
			energy=energy,
			verbose=0,
			//debug=strcat(prefix,"-debug-sc.dat"),
			//file  =strcat(prefix,"-field-sc.dat")
		);
		M = map_compose(SC,M);
	};
};

add_csr = {
	if(csr) {
		CSR = map_csr(psi,
			length=ls,
			angle=k0*ls,
			s=s_elem,
			ds=ds,
			energy=energy,
			filter=csr_filter,
			filterwidth=csr_filterwidth,
            transient=csr_transient,
            mode=csr_mode,
			verbose=0,
			//debug=strcat(prefix,"-debug-csr.dat"),
			//file  =strcat(prefix,"-field-csr.dat")
		);
		M = map_compose(CSR,M);
	};
};


element_loop = {
	lz = (z1-z0); // element length in z-coordinate
	if( 0<abs(k0) ) {
	  a0 = angle;
	  a1 = asin(sin(a0)-lz*k0);
	  ls = (a0-a1)/k0;
	} else {
		 ls = lz / cos(angle); // arc-length of element
	};
	dz = lz / N; // slice z-length; only approximately
	ds = ls / N; // slice arc-length

	s_elem = 0;              // arclength within element
	nelement = nelement + 1; // increment element number
	substep = 0;

	eval(hook_pre_element);

	do(N) {

        eval(hook_pre_step);

		// prefix for filenames
		prefix = format("%04g-%04g-", nelement, substep);
		prefix = strcat(prefix,elementname);

		// update variables
		z  =  z + dz;
		s  =  s + ds;
		s_elem = s_elem + ds;
		angle  = angle - k0*ds;

		eval( calc_dispersion );
		eval( calc_beamsize   );
		eval( calc_map        );
		eval( add_spacecharge );
		eval( propagation     );
		eval( output          );

		eval(hook_post_step);

		// increment step numbers
		substep = substep + 1;
		step    = step    + 1;

	};
	eval(hook_post_element);
};

// initialize output files
{} > "variance.dat";
{print("name  element  substep  step  z  s  ds  energy  gamma  betax  betay  beamsize  angle  R56  R56total");} >  "output.dat";

// output routine called after each step
output = {
	{ print(elementname, nelement, substep, step, z, s, ds, energy, gamma, betax, betay, beamsize, angle, R56, R56total); } >> "output.dat";
	if(dryrun==0) { { print(z, variance(psi)); } >> "variance.dat"; };
};


if(dryrun) {
	propagation={};
	add_csr={};
	add_spacecharge={};
};


// initialize global variables
z = 0;        // current position in floor coordinates
s = 0;        // current position in path length
nelement = 0; // number of current element
substep  = 0; // number of step within element
step     = 0; // total number of steps
angle    = 0; // cumulative angle
R = [         // current transfer matrix
1,0,0,0,
0,1,0,0,
0,0,1,0,
0,0,0,1
];

// hook defaults
hook_pre_step := '';
hook_post_step := {
    print(prefix);
	//print(psi);
	//show(psi, file=strcat(prefix,".ppm"), cells=0);
	//save(psi, strcat(prefix,".psd"));
	//write_localmoments(psi, strcat(prefix,"-localmoments.dat"), 1);
	//if(0.95 - integral(psi)) { print("integral too small, exiting"); save(psi,"failed.psd"); exit; };
};
hook_pre_element := '';
hook_post_element := '';
hook_begin := '';
hook_end := {
    { who(); } > 'variables.dat';
};

'''

	# VARIABLES
	for x in varlines:
		print(x,file=ofd)
	print('\n',file=ofd)

	# MACROS
	print(macrodefs,file=ofd)
	print('\n',file=ofd)

	# BETA DATA
	for x in betalines:
		print(x,file=ofd)
	print('\n',file=ofd)

	print("eval(hook_begin);\n\n",file=ofd)

	# LATTICE
	for x in latlines:
		print(x,file=ofd)

	print("eval(hook_end);\n",file=ofd)

if __name__ == '__main__':
	import datetime
	import sys
	import argparse

	version = '0.1.0'

	parser = argparse.ArgumentParser()
	parser.add_argument('lattice_file', help='file containing lattice data')
	parser.add_argument('beta_file',    help='file containing beta functions')
	parser.add_argument('-o', '--output',  help='output filename', default='')
	parser.add_argument('-v', '--version', action='version', version=version)
	args = parser.parse_args()

	if(args.output == ''):
		ofname = args.lattice_file
		ofname = ofname.replace('_lattice.dat','')
		ofname = ofname.replace('.lattice.dat','')
		ofname = ofname.replace('.dat','')
		ofname += '.inp'
	else:
		ofname = args.output



	header=f'''\
// generated by lattice2selav.py
// version {version}
// {datetime.datetime.now().isoformat()}
//
// lattice-file: {args.lattice_file}
// beta-file:  {args.beta_file}
//
'''
	with open(ofname,'w') as fd:
		print(header,file=fd,end='')
		with open(args.lattice_file,'r') as fdi:
			for line in fdi.readlines():
				print('// '+line,file=fd,end='')
			print('//\n',file=fd)

		lattice2selav(args.lattice_file, args.beta_file, fd)
