#!/usr/bin/env python3

import re
import numpy as np

def matchre(res,s,t=None):
    res = res.split(',')
    for r in res:
        s_match = re.search(r,s)
        t_match = re.search(r,t) if t else True
        if(s_match and t_match):
            return True
    return False

def renamere(res,s):
    res = res.split(',')
    for r in res:
        m = re.search(r,s)
        if(m):
            if m.groups():
                s = ''.join(m.groups())
            else:
                s = m.group()
    return s

def samematch(res,a,b):
    res = res.split(',')
    for r in res:
        ma = re.search(r,a)
        mb = re.search(r,b)
        if ma and mb and (ma.group() == mb.group()):
            return True
    return False

def mergeelements(elements,mergeres):

    def mergeable(a,b,mergeres=''):
        #if( a['name'] != b['name'] ):
        #    return False

        if( not samematch(mergeres, a['name'], b['name']) ):
            return False

        if( a['keyw'] != b['keyw'] ):
            return False

        for p in 'freq phase k0 k1'.split():
            if(p in a and a[p]!=b[p]):
                return False

        return True

    def merge(a,b):
        c=a.copy()
        #c['name']  = lcs(a['name'],b['name'])
        c['z1']  = b['z1']
        for p in 'deltae'.split():
            if p in c:
                c[p] += b[p]
        return c

    merged = []
    e_ = None
    for e in elements:
        if e_==None:
            e_=e.copy()
            continue
        if mergeable(e_,e,mergeres):
            e_ = merge(e_,e)
        else:
            merged.append(e_)
            e_ = e.copy()
    if e_:
        merged.append(e_)

    return merged


def renameelements(elements,res):
    for e in elements:
        e['name'] = renamere(res,e['name'])
    return elements

def dropelements(elements,res):
    return [ e for e in elements if not matchre(res,e['name']) ]

def make_family(elements):
    # count number of family members
    fmembers={}
    for e in elements:
        name = e['name']
        fmembers[name] = fmembers[name]+1 if name in fmembers else 1


    k0dict={}
    nfam={}
    for e in elements:
        name = e['name']
        if(e['keyw']=='dipole'):
            nfam[name] = nfam[name]+1 if name in nfam else 0
            if(e['name'] in k0dict):
                k0p = k0dict[e['name']] # parent k0
                if k0p!=0.0:
                    e['k0'] = (e['k0']/k0p)
                else:
                    # coefs for family sizes
                    # n-dipoles:  0   1      2        3          4       5?          6
                    fcoefs = [   [], [1], [1,-1], [1,-2,1], [1,-1,-1,1], [], [1,-1,-1,1,1,-1] ]
                    i=fmembers[name]
                    j=nfam[name]
                    if i<len(fcoefs) and j<len(fcoefs[i]):
                        e['k0'] = fcoefs[i][j]
                    else:
                        e['k0'] = 1
            else:
                k0dict[e['name']] = e['k0']
    return elements



latticeparams={
    'cavity':    'deltae freq phase'.split(),
    'dipole':    'k0 factor'.split(),
    'quadrupole':'k1'.split()
}

def formatelement(e):
    e = e.copy()

    fmt='{keyw:12s} {name:12s} {z0:9.4f} {z1:9.4f}'
    keywfmts={
        'cavity':'{deltae:14e} {freq:14e} {phase:14f}',
        'dipole':'{k0:+14g}',
        'quadrupole':'{k1:+14g}'
    }
    if e['keyw'] in keywfmts:
        fmt += ' ' + keywfmts[e['keyw']]
    return fmt.format(**e)

def write(elements,fname):
    with open(fname,"w") as fd:
        for e in elements:
            print(formatelement(e),file=fd)

def read(fname):
    elements = []
    with open(fname,"r") as fd:
        for line in fd.readlines():
            line = line.strip()
            if line[0]=='#':
                continue
            line=line.split()
            e={
                'keyw': line[0],
                'name': line[1],
                'z0': float(line[2]),
                'z1': float(line[3])
            }
            if(e['keyw'] in latticeparams):
                params = latticeparams[e['keyw']]
                for p in params:
                    e[p] = 0.0
                for p,l in zip(params,line[4:]):
                    e[p] = float(l)
            elements.append(e)
    return elements



if(__name__=='__main__'):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('lattice_file', nargs='*', default='/dev/stdin', help='lattice file')
    parser.add_argument('-d', '--drop', type=str, default='',
        help='comma separated list of regexes; matching elements will be dropped'
    )
    parser.add_argument('-r', '--rename', type=str, default='',
        help='comma separated list of regexes; matching elements will be renamed'
    )
    parser.add_argument('-m', '--merge', default='',
        help='merge elements yielding same match'
    )
    parser.add_argument('-f', '--family', default=False, action='store_true',
        help='make dipoles with same name a family'
    )
    parser.add_argument('-o', '--output', type=str, default='/dev/stdout',
        help='output file prefix'
    )
    args = parser.parse_args()


    # read data
    elements = read(args.lattice_file)

    # process elements
    if(args.drop):
        elements       = dropelements(elements, args.drop)
    if(args.merge):
        elements       = mergeelements(elements, args.merge)
    if(args.rename):
        elements       = renameelements(elements, args.rename)
    if(args.family):
        elements       = make_family(elements)

    write(elements,args.output)
