#!/usr/bin/env python3

import numpy as np

def readoptics(fname):
	elements={}
	repquo=lambda s: s.replace('"','')
	with open(fname,"r") as fd:
		head = fd.readline().strip()[1:].strip()
		formats = fd.readline().strip()[1:].strip().split()
		converters={i: repquo for i,fmt in enumerate(formats) if fmt[-1]=='s'}

		skip=2
		while True:
			line=fd.readline().strip()
			if( line[0]!='$' and line[0]!='@' ):
				break
			skip+=1
	data = np.genfromtxt(fname,
		skip_header=skip,
		dtype=None,
		encoding=None,
		names=head.split(),
		converters=converters
	)
	return data


def mad2elements(data, min_disp=1e-3, min_k0l=1e-3, min_k1l=1e-3):
	#
	#        s1p       s0     s1
	# ########.........########......
	#        |--- ds --|
	#			       |- ls -|
	#
	s1p = 0 # previous s1
	a0 = 0 # entrance angle
	z1 = 0 # current z position
	elements = []
	beta = []
	def get(e, key, default=0.0):
		return e[key] if key in e.dtype.names else default
	for elem in data:
		name   = elem['NAME']
		kw     = elem['KEYWORD']
		s1     = elem['S']
		ls     = elem['L'] if 'L' in elem else s1-s1p
		s0     = s1-ls
		betx   = elem['BETX']
		bety   = elem['BETY']
		disp   = elem['DX']
		# optional parameters not contained in traditional ll files
		k0l    = get(elem, 'K0L')
		k1l    = get(elem, 'K1L')
		deltae = get(elem, 'VOLT') * 1e6
		phase  = get(elem, 'LAG') * np.pi/180. # check unit
		freq   = get(elem, 'FREQ') * 1e6

		# calculate z-coordinates
		ds = s0-s1p
		a1 = a0+k0l # exit angle
		dz = ds * np.cos(a0) # inter-element space is always a drift
		lz = ls * np.abs(( (np.sin(a0)-np.sin(a1))/k0l if np.abs(k0l)>0.0 else np.cos(a0) ))

		# update angle and z coordinates
		a0 = a1
		z0 = z1 + dz
		z1 = z0 + lz
		s1p = s1

		# store beta functions
		beta.append([z1,betx,bety])

		e={'name':name, 'z0':z0, 'z1':z1}
		if(kw=='LCAVITY'):
			e['keyw']   = 'cavity'
			e['deltae'] = deltae
			e['freq']   = freq
			e['phase']  = phase

		elif(kw=='SBEND'):
			e['keyw']   = 'dipole'
			e['k0']     = k0l/ls

		elif(kw=='QUADRUPOLE') and (abs(disp)>=min_disp):
			e['keyw']   = 'quadrupole'
			e['k1']    = k1l/ls

		else:
			continue

		elements.append(e)

	return elements,beta


if(__name__=='__main__'):
	import latticetool
	import argparse
	import pathlib
	parser = argparse.ArgumentParser()
	parser.add_argument('mad_data_file', help='mad optics data')
	parser.add_argument('-o', '--output', type=str, default='',
		help='output file prefix'
	)
	args = parser.parse_args()

	# read data
	data = readoptics(args.mad_data_file)

	# process elements
	elements, beta = mad2elements(data)

	# determine output prefix
	if(args.output!=''):
		oprefix = args.output
	else:
		oprefix = pathlib.Path(args.mad_data_file).stem

	# write lattice data
	latticetool.write(elements,oprefix+'.lattice.dat')

	# write beta data
	zprev=None
	with open(oprefix+'.beta.dat',"w") as fd:
		for z,betax,betay in beta:
			if(z==zprev):
				continue
			print(f'{z:8.3f} {betax:8.3f} {betay:8.3f}',file=fd)
			zprev=z
